#!/usr/bin/perl -w

# Copyright 2013, 2014, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Print tikz lines which are the rows of the UArD tree low to high.  These
# go clockwise around and show the region of p,q coverage.

use strict;
use Math::PlanePath::PythagoreanTree;

use FindBin;
use lib $FindBin::Bin;
use MyPrintwrap ':all';
use MyTriples 'line_end_clip';

my $depth_limit = 3;
my $x_limit = 460;
my $y_limit = 460;
my $scale = 1;

my $coordinates = 'AB';
if (@ARGV) {
  $coordinates = shift @ARGV;
}
print "    % generated by tools/$FindBin::Script $coordinates\n";


sub is_outside {
  my ($x,$y) = @_;
  return ($x > $x_limit || $y > $y_limit);
};
sub fmt {
  my ($x) = @_;
  if ($x != int($x)) {
    return sprintf '%.2f', $x;
  } else {
    return $x;
  }
};
sub fmt_xy {
  my ($x,$y) = @_;
  return '('.fmt($x/$scale).','.fmt($y/$scale).')';
};

my $path = Math::PlanePath::PythagoreanTree->new(coordinates => $coordinates,
                                                 tree_type => 'UArD',
                                                 digit_order => 'LtoH');
{
  my ($x,$y) = $path->n_to_xy($path->n_start);
  print "    \\filldraw ", fmt_xy($x,$y), " circle(.25ex);\n";
}
$Printwrap = 4;
printwrap("\\draw");
foreach my $depth (1 .. 3) {
  my ($n_lo, $n_hi) = $path->tree_depth_to_n_range($depth);
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  unless ($depth==1) { printwrap_newline(); }
  $Printwrap = 6;
  printwrap(' '.fmt_xy($prev_x,$prev_y));
  $Printwrap = 8;
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);

    my $x1 = $prev_x;
    my $y1 = $prev_y;
    my $x2 = $x;
    my $y2 = $y;
    if (is_outside($x1,$y1) && is_outside($x2,$y2)) {
      ### both outside ...

    } elsif (is_outside($x2,$y2)) {
      ### from inside to outside the clip ...
      ($x2,$y2) = line_end_clip ($x1,$y1, $x2,$y2, $x_limit,$y_limit);
      printwrap ("--".fmt_xy($x2,$y2));

    } elsif (is_outside($x1,$y1)) {
      ### from outside to inside the clip ...
      ($x1,$y1) = line_end_clip ($x2,$y2, $x1,$y1, $x_limit,$y_limit);
      printwrap(" ".fmt_xy($x1,$y1)."--".fmt_xy($x2,$y2));

    } else {
      printwrap ("--".fmt_xy($x2,$y2));
    }
    $prev_x = $x;
    $prev_y = $y;
  }
}
printwrap(";\n");

print "    % end generated\n";
exit 0;
