#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl test.pl
#
# Run some self-tests on ../triples.tex and on MyTriples.pm.
#

use 5.010;
use strict;
use File::Slurp;
use IPC::Run;
use List::Util 'max','min';
use Math::Matrix;
use Math::NumSeq::OEIS::File;
use Regexp::Common 'no_defaults','balanced';

use FindBin;
use lib::abs $FindBin::Bin;
use MyTriples ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $source_dir = File::Spec->catdir ($FindBin::Bin, File::Spec->updir);
$source_dir = File::Spec->rel2abs($source_dir);
print "source dir $source_dir\n";

chdir $source_dir or die $!;

my $triples_tex_filename = 'triples.tex';
# $triples_tex_filename = 'devel/more.tex';
my $triples_pdf_filename = 'triples.pdf';
my $triples_tex_str = File::Slurp::slurp($triples_tex_filename);
print count_lines($triples_tex_str)," lines\n";

my $triples_tex_str_sans_comments = $triples_tex_str;
$triples_tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
print count_lines($triples_tex_str_sans_comments)," lines\n";


#------------------------------------------------------------------------------
# UAD iteration

{
  require Math::PlanePath::PythagoreanTree;
  my $path = Math::PlanePath::PythagoreanTree->new(coordinates => 'PQ',
                                                   tree_type => 'UAD');
  my $p = 2;
  my $q = 1;
  my $count = 0;
  foreach my $n ($path->n_start .. 500) {
    my ($p,$q) = $path->n_to_xy($n);
    my ($next_p,$next_q) = $path->n_to_xy($n+1);

    {
      my ($got_p,$got_q) = uad_pq_next($p,$q);
      unless ($got_p == $next_p && $got_q == $next_q) {
        die "UAD iteration n=$n";
      }
      $count++;
    }
    {
      my ($got_p,$got_q) = uad_pq_next_by_mdash($p,$q);
      unless ($got_p == $next_p && $got_q == $next_q) {
        die "UAD iteration n=$n";
      }
      $count++;
    }
  }
  print "UAD iterations $count\n";

  sub uad_pq_next {
    my ($p,$q) = @_;

    my $divisor = 2*$q;
    my $m = int(($p-$q) / $divisor);
    my $r = $p - $divisor*$m;
    ### uad_pq_next(): "$p,$q  m=$m r=$r"

    $p == $m*$divisor + $r or die;
    $r > $q or die;
    $r < $q+2*$q or die;

    if ($r == 2*$q) {
      ### U^(m+1) ...
      # $m++;
      # $r -= $divisor;
      # $m--;
      # $r += $divisor;
      # return (($m+2) * $r + -($m+3)*$q,
      #         ($m+1) * $r + -($m+2)*$q);

      # ($m+2) * $r + -($m+1)*$q      U^(m+1)
      # ($m+1) * $r + -($m)  *$q
      return ($m+3, $m+2);

    } elsif ($r < 2*$q) {
      ### reverse U ...
      return (-($m+1) * $r + (3*$m+4)*$q,
              -$m     * $r + (3*$m+1)*$q);
    } else {
      ### reverse A ...
      return (($m+2) * $r + -($m+3)*$q,
              ($m+1) * $r + -($m+2)*$q);
    }
  }

  sub uad_pq_next_by_mdash {
    my ($p,$q) = @_;

    my $divisor = 2*$q;
    my $m = int($p / $divisor);
    my $r = $p - $divisor*$m;

    $p == $m*$divisor + $r or die;
    $r >= 0 or die;
    $r < $divisor or die;

    if ($q == 1) {
      # ($m+2) * $r + -($m+1)*$q      U^(m+1)
      # ($m+1) * $r + -($m)  *$q
      return ($m+2, $m+1);

    } elsif ($r > $q) {
      ### reverse U ...
      return (-($m+1) * $r + (3*$m+4)*$q,
              -$m     * $r + (3*$m+1)*$q);
    } else {
      ### reverse A ...
      # ### uad_pq_next(): '5,2  m=0 r=5'
      return (($m+1) * $r + ($m)   * $q,
              ($m)   * $r + ($m-1) * $q);
    }
  }
}

#------------------------------------------------------------------------------
# check MY-BEGIN-3x3

{
  $triples_tex_str =~ /MY-BEGIN-3x3(.*?)MY-END-3x3/sg
    or die "MY-BEGIN-3x3 not matched";
  my $str = $1;
  my $linenum = pos_linenum($triples_tex_str);

  my $braces_re = $RE{balanced}{-parens=>'{}'};
  $str =~ s/\\vphantom$braces_re//g;

  $str =~ s/\\[hv]space\{[^}]*}//g;
  $str =~ s/%.*//g;
  $str =~ s!\{2\}!/2!g;
  $str =~ tr/{}/()/;
  $str =~ s/\\frac//g;
  $str =~ s/&/,/g;
  $str =~ s/\\\\(\[.*?\])?/],[/g;
  $str =~ s/\^/**/g;
  $str =~ s/([abcd])([abcd])/$1*$2/g;
  $str =~ s/([abcd])/\$$1/g;
  $str = "Math::Matrix->new([ $str ])";
  # print "$str\n";

  my $count = 0;
  foreach my $m (U_2x2(), A_2x2(), D_2x2(),
                 M1_2x2(), M2_2x2(), M3_2x2(),
                 T_2x2()) {
    my ($a,$b,$c,$d) = matrix_to_abcd($m);
    ### $str
    my $got_3x3 = eval "my \$a=$a; my \$b=$b; my \$c=$c; my \$d=$d; $str";
    my $want_3x3 = mm_2x2_to_3x3($m);
    $want_3x3->equal($got_3x3)
      or die "triples.tex:$linenum: 3x3 formula";
    $count++;
  }
  print "MY-BEGIN-3x3 ok $count\n";

  sub matrix_to_abcd {
    my ($m) = @_;
    return ($m->[0][0], $m->[0][1],
            $m->[1][0], $m->[1][1]);
  }
}

#------------------------------------------------------------------------------
# check U_3x3(),U_2x2() etc against PythagoreanTree path code
{
  require Math::PlanePath::PythagoreanTree;

  my $check_3x3 = sub {
    my ($tree_type, $m1,$m2,$m3) = @_;
    print "check_3x3() $tree_type\n";

    my $path1 = Math::PlanePath::PythagoreanTree->new(tree_type => $tree_type,
                                                      coordinates => 'AB');
    my $path2 = Math::PlanePath::PythagoreanTree->new(tree_type => $tree_type,
                                                      coordinates => 'AC');

    my $path_to_abc = sub {
      my ($n) = @_;
      my ($a, $b) = $path1->n_to_xy($n);
      (undef, my $c) = $path2->n_to_xy($n);
      return ($a,$b,$c);
    };

    foreach my $n (1 .. $path1->tree_depth_to_n_end(5)) {
      my ($a,$b,$c) = $path_to_abc->($n);
      my $v = Math::Matrix->new([$a],[$b],[$c]);

      my @children = $path1->tree_n_children($n);
      foreach my $m ($m1,$m2,$m3) {
        my $child_n = shift @children;
        my ($path_a,$path_b,$path_c) = $path_to_abc->($child_n);

        my $child_v = $m * $v;
        my $mult_a = $child_v->[0][0];
        my $mult_b = $child_v->[1][0];
        my $mult_c = $child_v->[2][0];
        unless ($mult_a == $path_a
                && $mult_b == $path_b
                && $mult_c == $path_c) {
          print "\nfrom n=$n  $a, $b, $c\n";
          print $child_v;
          print "mult $mult_a $mult_b $mult_c\n";
          print "path $path_a $path_b $path_c\n";
          print "bad n=$n in $tree_type rem=$#children\n";
          die;
        }
      }
    }
  };
  my $check_2x2 = sub {
    my ($tree_type, $m1,$m2,$m3) = @_;
    print "check_2x2() $tree_type\n";

    my $path = Math::PlanePath::PythagoreanTree->new(tree_type => $tree_type,
                                                     coordinates => 'PQ');

    foreach my $n (1 .. $path->tree_depth_to_n_end(5)) {
      my ($p,$q) = $path->n_to_xy($n);
      my $v = Math::Matrix->new([$p],[$q]);

      my @children = $path->tree_n_children($n);
      foreach my $m ($m1,$m2,$m3) {
        my $child_n = shift @children;
        my ($path_p,$path_q) = $path->n_to_xy($child_n);

        my $child_v = $m * $v;
        my $mult_p = $child_v->[0][0];
        my $mult_q = $child_v->[1][0];
        unless ($mult_p == $path_p
                && $mult_q == $path_q) {
          print "\nfrom n=$n  $p, $q";
          print $child_v;
          print "mult $mult_p $mult_q";
          print "path $path_p $path_q";
          print "bad n=$n in $tree_type rem=$#children\n";
          die;
        }
      }
    }
  };

  $check_3x3->('UAD', U_3x3(), A_3x3(), D_3x3());
  $check_3x3->('FB',  M1_3x3(),M2_3x3(), M3_3x3());
  $check_3x3->('UMT', U_3x3(), M2_3x3(), T_3x3());

  $check_2x2->('UAD', U_2x2(), A_2x2(), D_2x2());
  $check_2x2->('FB',  M1_2x2(),M2_2x2(), M3_2x2());
  $check_2x2->('UMT', U_2x2(), M2_2x2(), T_2x2());
}

#------------------------------------------------------------------------------
# UAD A,B in claimed regions

{
  require Math::PlanePath::PythagoreanTree;
  my $path = Math::PlanePath::PythagoreanTree->new;
  for (my $n = $path->tree_depth_to_n(1); $n < 50_000; $n += 3) {
    {
      my ($A,$B) = $path->n_to_xy ($n);
      die unless $A < 3/4 * $B;
    }
    {
      my ($A,$B) = $path->n_to_xy($n + 1);
      die unless $A > 3/4 * $B;
      die unless $A < 4/3 * $B;
    }
    {
      my ($A,$B) = $path->n_to_xy($n + 2);
      die unless $A > 4/3 * $B;
    }
  }
}
{
  require Math::PlanePath::PythagoreanTree;
  my $path = Math::PlanePath::PythagoreanTree->new (coordinates => 'PQ');
  for (my $n = $path->tree_depth_to_n(1); $n < 50_000; $n += 3) {
    {
      my ($p,$q) = $path->n_to_xy($n);
      die unless $p < 2 * $q;
    }
    {
      my ($p,$q) = $path->n_to_xy($n + 1);
      die unless $p > 2 * $q;
      die unless $p < 3 * $q;
    }
    {
      my ($p,$q) = $path->n_to_xy($n + 2);
      die unless $p > 3 * $q;
    }
  }
}

#------------------------------------------------------------------------------
# check 2x2 <-> 3x3 by formula

{
  my $check_3x3_2x2 = sub {
    my ($m_3x3, $m_2x2) = @_;
    print "check_3x3_2x2()  ",matrix_name($m_3x3),"\n";
    my $f_3x3 = mm_2x2_to_3x3($m_2x2);
    unless ($m_3x3->equal($f_3x3)) {
      die "oops";
    }
  };

  $check_3x3_2x2->(U_3x3(), U_2x2());
  $check_3x3_2x2->(A_3x3(), A_2x2());
  $check_3x3_2x2->(D_3x3(), D_2x2());

  $check_3x3_2x2->(M1_3x3(), M1_2x2());
  $check_3x3_2x2->(M2_3x3(), M2_2x2());
  $check_3x3_2x2->(M3_3x3(), M3_2x2());

  $check_3x3_2x2->(T_3x3(), T_2x2());

  # $t = mm_2x2_to_3x3(T_2x2());
  # $t = mm_2x2_to_3x3(T_2x2()->transpose)->transpose;
  # print $t;
}

#------------------------------------------------------------------------------
# check MY-BEGIN-MATRICES

{
  while ($triples_tex_str
         =~ /MY-BEGIN-MATRICES ([^\n]+)(.*?)MY-END-MATRICES/sg) {
    my $names = $1;
    my $content = $2;
    my $linenum = pos_linenum($triples_tex_str);
    my @names = split /,/, $names;

    my $i = 0;
    while ($content =~ /begin\{pmatrix\}(.*)?end\{pmatrix\}/g) {
      my $pmatrix = $1;
      ### $pmatrix
      ### $i

      my @got_terms = split m{[^-0-9/]+}, $pmatrix;
      if ($got_terms[0] eq '') { shift @got_terms } # leading empty

      # 1/2 -> 0.5
      foreach my $t (@got_terms) {
        if ($t =~ m{(.*)/2$}) {
          $t = $1 / 2;  # mutate array
        }
      }

      ### @got_terms
      my $got_terms = join(',',@got_terms);

      my $want_name = $names[$i%3];
      @got_terms == 4
        or @got_terms == 9
          or die "triples.tex:$linenum: MY-BEGIN-MATRICES $want_name terms: ",scalar(@got_terms),"\n";

      my $got_name = matrix_name(@got_terms);
      $got_name eq $want_name
        or die "triples.tex:$linenum: MY-BEGIN-MATRICES $got_terms want $want_name got $got_name\n";

      print "MY-BEGIN-MATRICES $want_name ok, line $linenum\n";
      $i++;
    }
  }
}

#------------------------------------------------------------------------------
# triples.pdf Title and Author

{

  my $pdfinfo_out;
  IPC::Run::run(['pdfinfo', '-isodates', $triples_pdf_filename],
                '>', \$pdfinfo_out)
      or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;
  {
    $pdfinfo_out =~ /^Title:\s*(.*)/m
      or die;
    my $title = $1;
    # print "title: $title\n";
    $title =~ /Trees/
      or die;
  }
  {
    $pdfinfo_out =~ /^Author:\s*(.*)/m
      or die;
    my $author = $1;
    # print "author: $author\n";
    $author =~ /Kevin/
      or die;
  }
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}
sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

#------------------------------------------------------------------------------
exit 0;
